"""Implements a Checker that makes use of OS X's QuickLook system."""
from pathlib import Path
from PIL import Image
import shutil
import subprocess
import tempfile
from spot_check_files.checker import Checker, CheckResult, CheckRequest


class QLChecker(Checker):
    def __str__(self):
        return 'QLChecker'

    """Checks files using QuickLook (MacOS only).

    This invokes the qlmanage command-line app and is fairly slow, but
    QuickLook can recognize a wide variety of file types and generates nice
    thumbnails.

    A file is assumed to be "valid" if QuickLook generates a thumbnail for it.
    If it doesn't, the checker will not mark the result as recognized.
    """
    def check(self, req: CheckRequest) -> CheckResult:
        result = CheckResult()

        outdir = tempfile.mkdtemp(dir=req.tmpdir)
        subprocess.check_output(
            ['qlmanage', '-t', '-s', '300',
             str(req.realpath), '-o', outdir])

        paths = list(Path(outdir).glob('*.png'))
        if paths:
            result.recognizer = self
            if req.thumb:
                with Image.open(str(paths[0])) as img:
                    img.load()
                    result.thumb = img
        else:
            result.errors.append('no png produced by qlmanage')

        shutil.rmtree(outdir)

        return result



"""Implementations of Checker for various basic file types."""
import csv
from importlib import resources
import json
from PIL import Image, ImageDraw, ImageFont, UnidentifiedImageError
from typing import List
from xml.dom import minidom
from spot_check_files import _monoid_font
from spot_check_files.checker import Checker, CheckRequest, CheckResult


_FONTS = []


def _font():
    if not _FONTS:
        with resources.path(_monoid_font, 'Monoid-Regular.ttf') as path:
            _FONTS.append(ImageFont.truetype(str(path), 8))
    return _FONTS[0]


def table_thumb(rows: List[List[str]]) -> bytes:
    """Builds a thumbnail for a table."""
    img = Image.new('L', (300, 300), color=255)
    draw = ImageDraw.Draw(img)
    font = _font()
    height = draw.textsize(' ')[1] + 5
    x = 0

    def rowcol(r, c):
        if c < len(r):
            return r[c]
        else:
            return ''

    for col in range(len(rows[0])):
        maxlen = min(max(len(rowcol(r, col)) for r in rows), 30)
        y = 0
        for row in rows:
            text = rowcol(row, col).replace('\n', '\\n')[0:maxlen]
            draw.text((x, y), text, fill=0, font=font)
            y += height
        x += draw.textsize(' ' * maxlen, font=font)[0] + 5
        draw.line([(x, 2), (x, y - 2)], fill=128)
        x += 5
    return img


def text_thumb(text: str) -> bytes:
    """Builds a thumbnail for the given text."""
    img = Image.new('L', (300, 300), color=255)
    draw = ImageDraw.Draw(img)
    draw.multiline_text((0, 0), text, fill=0, font=_font())
    return img


class CSVChecker(Checker):
    def __str__(self):
        return 'CSVChecker'

    """Checks a CSV or TSV file for problems.

    Python's csv.Sniffer is used to try to determine the column delimiter;
    if it can't figure it out, this checker will not mark the file as
    recognized.
    """
    def check(self, req: CheckRequest) -> CheckResult:
        result = CheckResult()
        try:
            rows = []
            with open(req.realpath, 'r', newline='') as file:
                dialect = csv.Sniffer().sniff(file.read(1024))
                result.recognizer = self
                file.seek(0)
                for row in csv.reader(file, dialect):
                    if req.thumb and len(rows) < 30:
                        rows.append(row[0:7])
                if req.thumb:
                    result.thumb = table_thumb(rows)
        except csv.Error as e:
            result.errors.append(e)
        return result


class ImageChecker(Checker):
    """Checks an image by loading it with PIL.
    """
    def __str__(self):
        return 'ImageChecker'

    def check(self, req: CheckRequest) -> CheckResult:
        result = CheckResult()
        try:
            with Image.open(req.realpath, 'r') as img:
                img.load()
                result.recognizer = self
                if req.thumb:
                    img.thumbnail((300, 300))
                    result.thumb = img
        except OSError as e:
            # An issue I've encountered but haven't dug into yet.
            result.errors.append(e)
        except UnidentifiedImageError as e:
            result.errors.append(e)
        except ZeroDivisionError as e:
            # An issue I've encountered but haven't dug into yet.
            result.errors.append(e)
        return result


class JSONChecker(Checker):
    """Checks that a file is valid JSON.

    If the file cannot be parsed, it will not be marked as recognized.
    """
    def __str__(self):
        return 'JSONChecker'

    def check(self, req: CheckRequest) -> CheckResult:
        result = CheckResult()
        parsed = None
        with req.realpath.open('r') as file:
            try:
                parsed = json.load(file)
                result.recognizer = self
            except json.JSONDecodeError as e:
                result.errors.append(e)
        if req.thumb and parsed is not None:
            pretty = json.dumps(parsed, indent=2)
            # Don't send an excessive amount of text to Pillow
            lines = [s[0:100] for s in pretty.splitlines()[0:100]]
            result.thumb = text_thumb('\n'.join(lines))
        return result


class PlaintextChecker(Checker):
    """Checks a plaintext file for encoding problems.

    This will read the input as a text file, relying on Python's default
    behavior to select the encoding. The checker will mark the file as
    recognized as long as it is able to open it. It will attempt to read
    the entire file and add an error if it cannot decode it.
    """
    def __str__(self):
        return 'PlaintextChecker'

    def check(self, req: CheckRequest) -> CheckResult:
        result = CheckResult()
        try:
            lines = []
            with open(req.realpath, 'r') as file:
                result.recognizer = self
                for line in file:
                    if req.thumb and len(lines) < 100:
                        lines.append(line[0:100])
                    pass
            if req.thumb:
                result.thumb = text_thumb(''.join(lines))
        except ValueError as e:
            result.errors.append(e)
        return result


class XMLChecker(Checker):
    """Checks that a file is valid XML.

    If the file cannot be parsed, it will not be marked as recognized.
    """
    def __str__(self):
        return 'XMLChecker'

    def check(self, req: CheckRequest) -> CheckResult:
        result = CheckResult()
        parsed = None
        try:
            parsed = minidom.parse(str(req.realpath))
            result.recognizer = self
        except Exception as e:
            result.errors.append(e)
        if req.thumb and parsed is not None:
            pretty = parsed.toprettyxml(indent='  ')
            # Don't send an excessive amount of text to Pillow
            lines = [s[0:100] for s in pretty.splitlines()[0:100]]
            result.thumb = text_thumb('\n'.join(lines))
        return result



"""Implementations of Checker for zip and tar files."""
from pathlib import Path
import tarfile
import tempfile
import zipfile
from zipfile import ZipFile
from spot_check_files.checker import Checker, CheckResult, CheckRequest


class ZipChecker(Checker):
    def __str__(self):
        return 'ZipChecker'

    """Extracts zip files."""
    def check(self, req: CheckRequest) -> CheckResult:
        result = CheckResult()

        try:
            if not zipfile.is_zipfile(req.realpath):
                result.errors.append('not a zipfile')
                return result

            result.recognizer = self

            with ZipFile(req.realpath, 'r') as zf:
                result.extracted = Path(tempfile.mkdtemp(dir=req.tmpdir))
                zf.extractall(result.extracted)
        except Exception as e:
            result.errors.append(e)

        return result


class TarChecker(Checker):
    def __str__(self):
        return 'TarChecker'

    """Extracts tar files.

    Tars may be compressed with any formats the python tarfile lib supports,
    including gz, bz2, and xz.
    """
    def check(self, req: CheckRequest) -> CheckResult:
        result = CheckResult()

        try:
            if not tarfile.is_tarfile(req.realpath):
                result.errors.append('not a tarfile')
                return result

            result.recognizer = self

            with tarfile.open(req.realpath, 'r') as tf:
                result.extracted = Path(tempfile.mkdtemp(dir=req.tmpdir))
                tf.extractall(result.extracted)
        except Exception as e:
            result.errors.append(e)

        return result



"""Defines the API for file checkers and a helper class for running them.

The Checker abstract class defines the interface that must be implemented
for each type of file that spotcheck will support. A Checker is given a
CheckRequest containing the path to a file and other info, and returns a
CheckResult indicating whether it knew how to interpret the file, any errors
it found in the file, and a thumbnail of the file. The Checker also extracts
files if the file is an archive (e.g. a zip or tar).

The CheckerRunner class recursively checks all the files in a given path
(including files inside archives, if it's configured with a Checker that
can extract the archive) using a given set of Checkers.
"""
from __future__ import annotations
from dataclasses import dataclass, field
from PIL import Image
from pathlib import Path
import platform
from tempfile import TemporaryDirectory
from typing import List, Union


@dataclass
class CheckRequest:
    """Represents a file that should be checked.

    Attributes:
        realpath - path to (possibly temporary) location of file
        tmpdir - a shared temporary directory
        virtpath - logical path associated with the file. For example,
                   if the file "foo/bar.txt" was extracted from "a.zip",
                   this might be "a.zip/foo/bar.txt".
        thumb - if True, a thumbnail should be generated
    """
    realpath: Path
    tmpdir: Path
    virtpath: Path
    thumb: bool = False


@dataclass
class CheckResult:
    """The result of a checking a file.
    Attributes:
        errors - any problems found with the file
        extracted - if the file was an archive, the path to the directory
                    containing its extracted contents. This should be a
                    subdirectory of the tmpdir specified in the request
        recognizer - a Checker will set this to itself if it's confident
                     that the file was valid or invalid; if the file type
                     is still unclear, this may be None
        thumb - a thumbnail of the file
    """
    errors: List[Union[str, Exception]] = field(default_factory=list)
    extracted: Path = None
    thumb: Image = None
    recognizer: Checker = None


class Checker:
    def check(self, req: CheckRequest) -> CheckResult:
        """Validate, extract, and/or make a thumbnail of the specified file.

        Errors relating to a problem with the file should be caught and added
        to the errors list of the result. Errors which may indicate a that the
        checker itself is unable to operate on valid files can be propagated.
        """
        raise NotImplementedError()


@dataclass
class FileSummary:
    """Records metadata about a file, and the result of checking it.

    Attributes:
        result - result from the first checker that recognized the file,
                 or an empty result
        size - size in bytes of the file
        virtpath - the logical path to the file (for files inside an archive,
                   this wil include the path to the archive)
    """
    size: int
    virtpath: Path
    result: CheckResult = field(default_factory=CheckResult)


class CheckerRunner:
    """This is the main entry point for spot-checking files.

    What types of files will be checked, and how they will be recognized,
    is determined by the checkers attribute. The CheckerRunner.default()
    method will create an instance with a default set of checkers.

    The check_path method should be called for each file or directory you
    wish to check.

    Attributes:
        checkers - for each file, these will be applied in order until
                    one recognizes the file
    """
    @classmethod
    def default(cls):
        """Returns an instance with hopefully-reasonable defaults.

        Default checkers, in order:
            1. FileNameChecker.default()
            2. If on a Mac, QlChecker
        """
        from spot_check_files.filenames import FileNameChecker
        checkers = [FileNameChecker.default()]
        if platform.mac_ver()[0]:
            from spot_check_files.quicklook import QLChecker
            checkers.append(QLChecker())
        return cls(checkers)

    def __init__(self, checkers: List[Checker] = []):
        self.checkers = list(checkers)

    def check_path(self, path: Path, virtpath: Path = None,
                   tmpdir: Path = None) -> List[FileSummary]:
        """Runs checks against the file or directory at the given path."""
        if not tmpdir:
            with TemporaryDirectory() as tmpdir:
                return self.check_path(path, virtpath, Path(tmpdir))
        results = []
        virtpath = virtpath or path
        if path.is_dir():
            for fpath in path.glob('**/*'):
                if fpath.is_file():
                    fvirtpath = virtpath.joinpath(fpath.relative_to(path))
                    results.extend(self.check_path(fpath, fvirtpath, tmpdir))
            return results
        elif path.is_file():
            summary = FileSummary(
                virtpath=virtpath,
                size=path.stat().st_size)
            results.append(summary)
            for checker in self.checkers:
                req = CheckRequest(
                    realpath=path, tmpdir=tmpdir, virtpath=virtpath,
                    thumb=True)
                res = checker.check(req)
                if res.recognizer:
                    summary.result = res
                    if res.extracted:
                        results.extend(
                            self.check_path(res.extracted, virtpath, tmpdir))
                    break
        return results



"""Implements the spotcheck command-line tool."""
import argparse
from pathlib import Path
from typing import List
from spot_check_files.checker import CheckerRunner
from spot_check_files.report import CheckReport


def main(args: List[str] = None) -> int:
    """Checks one or more paths and displays the output.

    See docs/usage.txt for usage.

    If args is omitted, the process's command-line args are used.

    Returns 0 to indicate success.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('path', nargs='+', help='file or folders to check')
    parser.add_argument('-H', '--html', action='store_true', default=False,
                        help='output HTML')
    args = parser.parse_args(args)
    runner = CheckerRunner.default()
    summaries = []
    for path in args.path:
        summaries.extend(runner.check_path(Path(path)))
    report = CheckReport(summaries)
    if args.html:
        print(report.html())
    else:
        report.print()
    return 0



"""Handles summarizing and formatting results. Main class is CheckReport."""
import base64
from io import BytesIO
import os
from PIL import Image
from terminaltables import AsciiTable
from typing import List
from spot_check_files.checker import FileSummary


def _print_thumbs(summaries):
    try:
        from imgcat import imgcat
    except ImportError:
        return

    while summaries:
        # Combine images to show three per line in the output
        group = [s.result.thumb for s in summaries[0:3]]
        boxes = [i.getbbox() for i in group]
        if any(boxes):
            width = sum(b[2] - b[0] + 1 for b in boxes if b)
            height = max(b[3] - b[1] + 1 for b in boxes if b)
            img = Image.new('RGBA', (width, height))
            x = 0
            for i in range(len(group)):
                if boxes[i] is None:
                    continue
                img.paste(group[i], (x, 0))
                x += boxes[i][2] - boxes[i][0] + 1
            imgcat(img)
        summaries = summaries[3:]


class _GroupStats:
    def __init__(self, name: str, summaries: List[FileSummary],
                 comparison: List[FileSummary] = None):
        self.name = name
        self.count = len(summaries)
        self.size = sum(s.size for s in summaries)
        self.count_pct = ''
        self.size_pct = ''
        if comparison:
            self.count_pct = '{:.0%}'.format(self.count / len(comparison))
            self.size_pct = '{:.0%}'.format(
                self.size / sum(s.size for s in comparison))


class CheckReport:
    """Formats and displays results from a CheckerRunner."""
    def __init__(self, summaries: List[FileSummary]):
        self.summaries = summaries
        arch_summaries = [s for s in summaries
                          if s.result.extracted
                          and not s.result.errors]
        leaf_summaries = [s for s in summaries
                          if s.result.errors or not s.result.extracted]
        self.err_summaries = [s for s in leaf_summaries if s.result.errors]
        self.thumb_summaries = [s for s in leaf_summaries if s.result.thumb]
        self.groups = [
            _GroupStats('Archives (without errors)', arch_summaries),
            _GroupStats('Files (excludes errorless archives)',
                        leaf_summaries),
            _GroupStats('Files with thumbnails', self.thumb_summaries,
                        leaf_summaries),
            _GroupStats('Files with ERRORS', self.err_summaries,
                        leaf_summaries),
        ]

        by_rec = {}
        unrec_by_ext = {}
        for summary in leaf_summaries:
            if summary.result.recognizer is None:
                ext = summary.virtpath.suffix
                if ext not in unrec_by_ext:
                    unrec_by_ext[ext] = []
                unrec_by_ext[ext].append(summary)
            else:
                rec = str(summary.result.recognizer)
                if rec not in by_rec:
                    by_rec[rec] = []
                by_rec[rec].append(summary)

        for rec in sorted(by_rec.keys()):
            self.groups.append(_GroupStats(f'Recognized by {rec}',
                                           by_rec[rec], leaf_summaries))

        for ext in sorted(unrec_by_ext.keys()):
            if not ext:
                continue
            self.groups.append(_GroupStats(
                f'Unrecognized with extension {ext}',
                unrec_by_ext[ext], leaf_summaries))
        if '' in unrec_by_ext:
            self.groups.append(_GroupStats(
                'Other unrecognized', unrec_by_ext[''], leaf_summaries))

    def print(self):
        """Prints the report to the terminal.

        If using iTerm2, thumbnails are displayed; otherwise, they are
        not used.
        """
        if os.environ.get('TERM_PROGRAM', None) == 'iTerm.app':
            _print_thumbs(self.thumb_summaries)

        for summary in self.err_summaries:
            print(f'ERRORS for {summary.virtpath}')
            for err in summary.result.errors:
                print(f'\t{err}')
            print()

        stats = [('Group', 'Files', 'Size', '% Files', '% Size')]
        stats.extend((g.name,
                      g.count,
                      g.size,
                      g.count_pct,
                      g.size_pct) for g in self.groups)
        table = AsciiTable(stats)
        for i in range(1, 5):
            table.justify_columns[i] = 'right'
        print(table.table)

    def html(self) -> str:
        """Returns an HTML version of the report, as a string."""
        from jinja2 import Environment, PackageLoader, select_autoescape

        def thumburl(summary):
            data = BytesIO()
            summary.result.thumb.save(data, 'png')
            encdata = base64.b64encode(data.getvalue()).decode('utf-8')
            return f'data:image/png;base64,{encdata}'

        env = Environment(
            loader=PackageLoader('spot_check_files', '_templates'),
            autoescape=select_autoescape(['html']),
            trim_blocks=True)
        env.globals['thumburl'] = thumburl
        template = env.get_template('report.html')
        return template.render(vars(self))



"""Logic for selecting a Checker based on filenames (e.g. file extension)."""
from __future__ import annotations
import fnmatch
from typing import List, Tuple
from spot_check_files.archives import TarChecker, ZipChecker
from spot_check_files.basics import CSVChecker, ImageChecker,\
    JSONChecker, PlaintextChecker, XMLChecker
from spot_check_files.checker import Checker, CheckResult, CheckRequest


class FileNameChecker(Checker):
    """Selects and runs a Checker based on a file's (logical) path.

    If the Checker matching the filename does not recognize the file,
    this Checker will mark itself as the recognizer and add an error.

    The virtpath attribute of the CheckRequest is used.

    Attributes:
        checkers - list of tuples mapping filename pattern
                   (as specified in fnmatch module) to Checker
    """

    @classmethod
    def default(cls) -> FileNameChecker:
        """Returns an instance configured for some standard file types."""
        ctar = TarChecker()
        checkers = [
            ('*.csv', CSVChecker()),
            ('*.json', JSONChecker()),
            ('*.md', PlaintextChecker()),
            ('*.tar', ctar),
            ('*.tar.bz2', ctar),
            ('*.tar.gz', ctar),
            ('*.tar.xz', ctar),
            ('*.tbz', ctar),
            ('*.tgz', ctar),
            ('*.txt', PlaintextChecker()),
            ('*.txz', ctar),
            ('*.xml', XMLChecker()),
            ('*.zip', ZipChecker()),
        ]

        # I haven't tested most of these, they're just some of the ones
        # that Pillow supports that seem most likely to be of use to me
        img_exts = [
            'bmp',
            'gif',
            'icns',
            'ico',
            'jpg',
            'jpeg',
            'png',
            'tiff',
            'webp',
        ]
        cimg = ImageChecker()
        checkers.extend((f'*.{ext}', cimg) for ext in img_exts)

        return cls(checkers)

    def __str__(self):
        # TODO it's bad that this doesn't include any details about
        #      the instance. Possibly I shouldn't be using __str__
        #      in the reports, so that there would be less pressure
        #      for this to be short/readable. Ideally even the report
        #      would print some way of distinguishing instances, though.
        return 'FileNameChecker'

    def __init__(self, checkers: List[Tuple[str, Checker]] = []):
        self.checkers = list(checkers)

    def check(self, req: CheckRequest) -> CheckResult:
        for (pattern, checker) in self.checkers:
            if fnmatch.fnmatch(str(req.virtpath), pattern):
                result = checker.check(req)
                if result.recognizer is None:
                    result.recognizer = self
                    result.errors.append(
                        f'expected to be recognized by {checker} because '
                        f'filename matched: {pattern}')
                return result
        return CheckResult()



"""A tool for validating and gathering statistics on a set of files.

The cli module implements the command-line interface. It inspects files
using checker.CheckerRunner (which uses Checkers from the archives,
basics, filenames, and quicklook modules) and displays results using
report.CheckReport.
"""


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="spot_check_files",
    version="0.0.1",
    author="Jacob Williams",
    author_email="jacobaw@gmail.com",
    description="Helps validate the integrity of data backups/exports.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/brokensandals/spot_check_files",
    packages=setuptools.find_packages('src'),
    package_data={
        'spot_check_files._monoid_font': ['Monoid-Regular.ttf'],
        'spot_check_files._templates': ['report.html'],
    },
    package_dir={'': 'src'},
    entry_points={
        'console_scripts': [
            'spotcheck = spot_check_files.cli:main'
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS :: MacOS X",
    ],
    install_requires=[
        'Jinja2',
        'Pillow',
        'terminaltables',
    ],
    extras_require={
        'imgcat': ['imgcat'],
    },
    python_requires='>=3.7',
)
